/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include "etherxd.h"

/* code for loading registry */

xmlnode config = NULL;

void registry_load(char *regfile)
{
    xmlnode_free(config); /* in case this is a reload */

    config = xmlnode_file(regfile);
    if (config == NULL)
    {
        fprintf(stderr,"Unable to load configuration file: %s\nUsage: etherxd -r /path/to/registry.xml\nExiting...\n\n",regfile);
        exit(0);
    }
    jdebug(ZONE,"Loaded registry %s.", regfile);
    registry_include(config);
}

/* recursive include checker */
void registry_include(xmlnode x)
{
    xmlnode next, load;
    static int sane = 0;

    ++sane;
    if(sane > 100)
    {
        fprintf(stderr,"Your registry is recursively including itself, please correct it.\nExiting...\n\n");
        exit(0);
    }

    next = xmlnode_get_firstchild(x);
    while(next != NULL)
    {
        if(j_strcmp(xmlnode_get_name(next),"include") == 0)
        {
            jdebug(ZONE,"registry including %s",xmlnode_get_data(xmlnode_get_firstchild(next)));
            load = xmlnode_file(xmlnode_get_data(xmlnode_get_firstchild(next)));
            xmlnode_insert_node(x,load);
            xmlnode_free(load);
            xmlnode_hide(next);
            next = xmlnode_get_firstchild(x); /* start over, rescan after we include the new file */
            continue;
        }

        if(xmlnode_get_type(next) == NTYPE_TAG)
            registry_include(next);

        next = xmlnode_get_nextsibling(next);
    }

    --sane;
}


struct registry_t *registry_receivers = NULL;

void registry_handle(char *id, char *ns, void (*etherx_receive)(epacket e), thread t)
{
    struct registry_t *r;
    xmlnode x = NULL;

    if(id == NULL || ns == NULL)
        return;

    x = xmlnode_get_firstchild(config);
    while(x != NULL)
    {
        if( j_strcmp(xmlnode_get_name(x),"transport") == 0 && 
            j_strcmp(xmlnode_get_attrib(x,"id"),id) == 0 && 
            j_strcmp(xmlnode_get_attrib(x,"namespace"),ns) == 0
          ) break;
        x = xmlnode_get_nextsibling(x);
    }

    r = registry_lookup(id,ns);
    if(r == NULL)
    {
        r = malloc(sizeof(struct registry_t));
        r->id = strdup(id);
        r->ns = strdup(ns);
        r->reg = NULL;
        r->next = registry_receivers;
        registry_receivers = r;
    }

    /* always reset these */
    r->t = t;
    r->etherx_receive = etherx_receive;

    registry_send(r,x);
}

void registry_send(struct registry_t *r, xmlnode x)
{
    if(r == NULL)
        return;

    jdebug(ZONE,"sending config data");

    if(r->reg != NULL)
    {
        /* this is a reconfigure, and their registry didn't change, skip */
        if(xmlnode_cmp(r->reg,x) == 0)
            return;
    }else{
        r->reg = xmlnode_dup(x);
    }

    if(x == NULL)
        x = xmlnode_new_tag("etherx:registry");
    else
        x = xmlnode_dup(x);

    /* HACKERY, but it works nicely... break in and change the name */
    x->name = pstrdup(x->p, "etherx:registry");

    if(r->t != NULL)
        xmlstream_write(r->t,x);
    if(r->etherx_receive != NULL)
        (r->etherx_receive)(epacket_new(x,NULL,NULL,NULL));
}

void registry_scan()
{
    struct registry_t *r = registry_receivers;
    xmlnode next;

    next = xmlnode_get_firstchild(config);
    while(next != NULL)
    {
        /* resend to existing transports */
        if( j_strcmp(xmlnode_get_name(next),"transport") == 0 && 
            xmlnode_get_attrib(next,"id") != NULL && 
            xmlnode_get_attrib(next,"namespace") != NULL && 
            (r = registry_lookup(xmlnode_get_attrib(next,"id"),xmlnode_get_attrib(next,"namespace"))) != NULL
          ) registry_send(r, next);

        /* load new dso's */ 
        if( j_strcmp(xmlnode_get_name(next),"transport") == 0 && 
            xmlnode_get_attrib(next,"dso") != NULL && 
            xmlnode_get_attrib(next,"id") != NULL && 
            xmlnode_get_attrib(next,"namespace") != NULL && 
            registry_lookup(xmlnode_get_attrib(next,"id"),xmlnode_get_attrib(next,"namespace")) == NULL
          ) dso_load(next);

        next = xmlnode_get_nextsibling(next);
    }
}

struct registry_t *registry_lookup(char *id, char *ns)
{
    struct registry_t *r = registry_receivers;

    if(id == NULL || ns == NULL)
        return NULL;

    while(r != NULL)
    {
        if(j_strcmp(r->id,id) == 0 && j_strcmp(r->ns,ns) == 0)
            return r;
        r = r->next;
    }

    return NULL;
}