/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include "etherxd.h"

struct transport_dblist
{
    edb d;
    struct transport_dblist *next;
};

struct transport_data
{
    char *namespace;
    char *id;
    struct transport_dblist *d;
};

/* this is the callback for socket transports */
void transport_onNode(thread t, xmlnode in, int type)
{
    xmlnode x;
    tstream ts = (tstream)(t->data);
    struct transport_data *td = (struct transport_data *)(ts->data_aux5);
    struct transport_dblist *td_d;
    epacket e;
    edb d;

    /* we need to
     * - remember the namespace for all the epackets we create
     * - authenticate transport for EPACKET_TRANSPORT
     * - create db entries for EPACKET_ALIAS
     * - send registry for EPACKET_REGISTRY
     * - hide etherx:to, etherx:from attribs
     * 
     */

    jdebug(ZONE,"transport: thread %d and type %d",t,type);

    if(type == XMLSTREAM_ROOT)
    {
        td = (struct transport_data *)pmalloc(t->p, sizeof(struct transport_data));
        ts->data_aux4 = (void *)0;
        ts->data_aux5 = (void *)td;
        td->namespace = pstrdup(t->p, xmlnode_get_attrib(in, "xmlns"));
        td->id = NULL;
        td->d = NULL;
        pth_cleanup_push(transport_cleanup, (void *)t);
        return;
    }

    if(type == XMLSTREAM_ERROR)
    {
        jerror("etherxd","socket connected transport sent illegal XML");
        return;
    }

    if(type == XMLSTREAM_CLOSE)
    {
        jerror("etherxd","socket connected transport closed connection");
        return;
    }

    if(type == XMLSTREAM_BOUNCE)
    {
        e = epacket_new(xmlnode_dup(in), NULL, NULL, NULL);
        jerror("etherxd","bouncing data intended for trasport %s",e->to);
        epacket_error(e,"delivery failure");
        manager_send(e);
        return;
    }

    x = xmlnode_dup(in);
    e = epacket_new(x, NULL, NULL, td->namespace);

    jdebug(ZONE,"transport-e: %d",e->type);

    switch(e->type)
    {
        case EPACKET_NORMAL:
            manager_send(e);
            ts->data_aux4++;
            return;
        case EPACKET_TRANSPORT:
            if(transport_secret(xmlnode_get_data(xmlnode_get_firstchild(x)),xmlnode_get_attrib(x,"secret")))
            {
                td->id = pstrdup(t->p, xmlnode_get_data(xmlnode_get_firstchild(x)));
                xmlnode_hide_attrib(x,"secret");
                xmlstream_write(t,x);
                return;
            }else{
                xmlstream_error(t,"invalid");
            }
            break;
        case EPACKET_ALIAS:
            if(td->id == NULL)
                break;
            /* register this thread for this host/namespace */
            d = db(CMD_SET, xmlnode_get_data(xmlnode_get_firstchild(x)), td->namespace);
            d->t = t;
            d->type = DB_TRANSPORT;
            /* save this db entry on our own list of them so we can later flag them if we dissappear */
            td_d = (struct transport_dblist *)pmalloc(t->p, sizeof(struct transport_dblist));
            td_d->d = d;
            td_d->next = td->d;
            td->d = td_d;
            jdebug(ZONE,"got etherx:alias=%s",d->host);
            break;
        case EPACKET_REGISTRY:
            if(td->id == NULL || j_strcmp(xmlnode_get_attrib(x,"type"),"request") != 0)
                break;

            /* they asked for their config, give it to em :) */
            jdebug(ZONE,"got etherx:registry request");
            registry_handle(td->id, td->namespace, NULL, t);
            break;
    }
    xmlnode_free(x);
}

int transport_secret(char *id, char *secret)
{
    xmlnode x;

    x = xmlnode_get_firstchild(config);
    while(x != NULL)
    {
        if( j_strcmp(xmlnode_get_name(x),"transport") == 0 &&
            j_strcmp(xmlnode_get_attrib(x,"id"),id) == 0 &&
            j_strcmp(xmlnode_get_attrib(x,"secret"),secret) == 0
          ) return 1;
        x = xmlnode_get_nextsibling(x);
    }
    return 0;
}

void transport_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    struct transport_data *td = (struct transport_data *)(ts->data_aux5);
    struct transport_dblist *next = td->d;

    jdebug(ZONE,"cleaning up transports db entries");

    while(next != NULL)
    {
        next->d->type = DB_ERROR;
        logger_transport2(td->namespace, next->d->host, td->id, next->d->count);
        next = next->next;
    }

    if(td->id != NULL)
        logger_transport(td->namespace, ts->host, td->id, (int)ts->data_aux4);
    registry_handle(td->id, td->namespace, NULL, NULL); /* turn it off so a reconfigure doesn't go nowhere */
}

/* all this thing does is take data from an incoming xmlstream associated with a transport and feed it to it */
void transport_deliver_onNode(thread t, xmlnode in, int type)
{
    tstream ts = (tstream)(t->data);
    edb d = (edb)(ts->data_aux5);

    jdebug(ZONE,"connection->transport");

    if(type == XMLSTREAM_ROOT)
    {
        ts->data_aux4 = (void *)0;
        pth_cleanup_push(transport_deliver_cleanup, (void *)t);
        return;
    }

    if(d->type == DB_ERROR)
    {
        jwarn("etherxd","recipient transport disconnected, dropping connection");
        xmlstream_error(t, "recipient closed connection");
    }

    if(type == XMLSTREAM_NODE)
    {
        manager_deliver(d, epacket_new(xmlnode_dup(in), d->host, ts->host, NULL));
        ts->data_aux4++;
    }
}

void transport_deliver_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    edb d = (edb)(ts->data_aux5);

    logger_conn(d->namespace, ts->host, d->host, (int)ts->data_aux4);
}

